#exec(open('def_read_line_transitions.py').read())
#home_dir = os.environ['HOME']
#ion_dir = os.path.join(home_dir,'Dropbox','VoigtFit', 'static', '')
ion_dict = {'OI_1302':0, 'NI_1199':1, 'NI_1200':2, 'NI_1200.7':3, 'SII_1250':4, 'SII_1253':5, 'SII_1259':6,\
             'NiII_1317':7, 'NiII_1370':8, 'CII_1334':9, 'CIIx_1335':10,'FeII_1144':11, 'FeII_1608':12, 'AlII_1670':13,\
             'PII_1152':14, 'SiII_1190':15, 'SiII_1193':16,'SiII_1260':17,'SiII_1304':18,'SiII_1526':19,'SiIII_1206':20,\
             'SiIV_1393':21, 'SiIV_1402':22, 'CIV_1548':23, 'CIV_1550':24, 'NV_1238':25, 'NV_1242':26,'OVI_1031':27, 'OVI_1037':28 
            }
#add SiII 1304, PII1152 to the ion dictionary
#OI-1302 SII-1253 NI-1199 SiII-1190 PII-1152 FeII-1144 \
#              NI-1134 FeII-1125 ArI-1048 OI-1039 SiII-1020 \
#              OI-988a  NI-972  PII-963  NI-953  HI-949  OI-948a
import numpy as np
def read_line_transitions(ion_dir, ion_dict):
    """
    Reads in the line transitions, central wavelengths, and oscillator strengths for the given ions in the ion_dict from the linelist.dat
    file located under VoigtFit package located in the Dropbox directory. There are multiple linelist.dat files under the main VoigtFit directory.
    We want to use the linelist.dat in the main Voigtfit directory in static. 
    ----------
    ion_dir : This is the location path to the linelist.dat file
    ion_dict :this is a hard coded dictionary containing all of the line transitions you are interested in plotting

    Defaults
    ---------
    This code is specifically hard coded to read in the linelist.dat file in the static directory under the 1st Voigtfit directory
    given a specific ion dictionary. 

    Returns
    -------
    central_wl_ion_dict and ion_f_value_dict: Two dictionaries enclosed in a list. central_wl_ion_dict hold the central wavelengths
    of the line transitions as the keys and the string name of the line transitions as the values. ion_f_value_dict holds the string name 
    of the line transitions as the keys and the oscillator strengths as values. To access each dictionary by itself use the following notation:
    read_line_transitions(ion_dir, ion_dict)[0] to return central_wl_ion_dict
    and read_line_transitions(ion_dir, ion_dict)[1] to return ion_f_value_dict
    
    Example on How to Run Code
    --------------------------
    read_line_transitions(ion_dir,ion_dict)[0] returns the first dictionary (central_wl_ion_dict)
    
    Written by April Horton 4/10/2022
    """
    line_list_file  = open(ion_dir+'linelist.dat', "r")
    line_list_data = line_list_file.readlines()[7:]#start reading at index 7 because of the headers, this should be standard
    #for all users because we are using the same linelist.dat
    ions = []
    central_wl = []
    oscillator = []
    #ion_dict is a hard coded dictionary. If you want to add additional line transitions to the plots stacks, they will need to be added by hand. 
    for line in line_list_data: #iterate through each of the lines in the data file
        line = line.strip() #removes spaces at the beginning and at the end of the string. Returns a copy of the string with both leading and trailing characters removed
        if not line or line.startswith("#"): #if the line is blank or is a comment, starts with #
            continue #ends the current iteration and goes to the next one
   
        columns = line.split() # line.split() splits the line of strings into a list of strings.     
        col_id = columns[0] #holds the line transition string name
        ions.append(col_id)
        cen_wl = columns[2] #holds the central wavelengths of the transitions from the linelist.dat file
        central_wl.append(cen_wl)
        f_value = columns[3] #holds the oscillator values of the transitions 
        oscillator.append(f_value)
    line_list_file.close()

    ions_array=np.array(ions) #convert the ions list to an array for numpy functionality 
    keys = [] #these keys will be the central wavelengths
    ion_strings = [] #the string name of the line transitions
    oscillator_strengths = [] #holds the oscillator strengths for each line transition
    for i in ion_dict: #loop through all of the ions inside the ion_dict
        if i in ions_array: #is the ion actually in the line_list.dat file, helps for human mistakes when creating ion_dict
            if np.count_nonzero(ions_array==i)>1: #if there are duplicate ions at the same wavelength, compare oscillator strengths to find the ion with the larger f_value
                indices = np.where(ions_array==i)[0] #find the indices where duplicate ions occur
                f_value_list = [] #stores the oscillator values of the duplicate ions
                for x in range(len(indices)): #iterate over the indices to grab the oscillator values for comparison
                    f_value_list.append(float(oscillator[indices[x]]))
                #then, I find the maximum value in the oscillator strength list and return that index. I use that index in the indices list to grab the right element frm the ions, central wl and oscillator lists.
                keys.append(float(central_wl[indices[f_value_list.index(max(f_value_list))]]))
                ion_strings.append(ions[indices[f_value_list.index(max(f_value_list))]])
                oscillator_strengths.append(float(oscillator[indices[f_value_list.index(max(f_value_list))]]))
            else:
                keys.append(float(central_wl[ions.index(i)]))
                ion_strings.append(ions[ions.index(i)])
                oscillator_strengths.append(float(oscillator[ions.index(i)]))
            #I use the ions.index(i) because i is a string of the ion's name. Therefore, I need the index from the ions list where that ion string occurs
        else:
            return("Sorry, ion "+i+" is not in line_list.dat")

    #the zip function essentially works as a zipper. It "zips" together the first item with the second item. It joins the two iterators together
    x = zip(keys, ion_strings)
    central_wl_ion_dict = dict(x) #this dictionary has the central wavelengths as the keys and the line transition strings as the value

    y = zip(ion_strings, oscillator_strengths)
    ion_f_value_dict = dict(y) #this dictionary holds the line transitions as the keys and the oscillator strengths as the values
    return [central_wl_ion_dict, ion_f_value_dict]
#ion_dict = ['CII_1334']
#rest_wl = list(read_line_transitions(ion_dir,ion_dict)[0].keys())[0] #getting just the laboratory frame rest wavelength
#print(rest_wl)





#import numpy as np 

#exec(open('fitting.py').read())
def best_fit(x, y, w, deg = 8, epsilon = 0.1,force_degree=False):

	dat = []
	if not force_degree: #not forcing the degree of polynomial to be 8
		for i in range(deg):
			z, SS_res, rank, sing, rc = np.polyfit(x, y,i+1,full=True,w=w) #np.polyfit (x-coordinates, y-coordinates, polynomial order, Full=True means that the 
				#diagnostic information is returned )
			#SS_res = the residuals from the fit

			p            = np.poly1d(z) #z is the polynomial coefficients
			sorted_x     = np.sort(x)
			fit_y        = p(sorted_x) #fitting the polynomial to the x coordinates
			dof          = len(x)-len(z)
			numerator 	 = np.sum(((y-fit_y)/w)**2)

			#numerator    = SS_res
			#denominator  = sum(np.array(fit_y)**2)
			#fraction     = numerator/denominator
			red_chi_squared  = numerator/dof
			#SS_res is the sum of the squared residuals
			#SS_tot       = np.sum((y - np.mean(y))**2)
			#r_squared    = 1 - ((SS_res / SS_tot) * ((len(y) - 1) 
			#	/ (len(y) - dof - 1)) )

			polydegree   = {'xarray': x,
						    'fit_y': fit_y,
						    'red_chi_squared': red_chi_squared,
						    'degree': i+1}
			
			if i >= 1:
				if abs(polydegree['red_chi_squared'] - dat[i-1]['red_chi_squared']) < epsilon: #it is a good fit and the difference is less 
				#than the epsilon value. Might as well keep the fit before it
					#print(polydegree['red_chi_squared'], dat[i-1]['red_chi_squared'] )
					#print("This is a good fit of order " +str(i+1))

					return polydegree['degree']
					break #as soon as the program finds a fit that is less than epsilon it breaks and no longer needs to continue 
				else: dat.append(polydegree) #if the current fit is not less than epsilon 
			else: dat.append(polydegree) #this is for the very first time through the degree values. It creates a starting point to compare 
			#the future polynomial fits to. 
	else: #if you specifically want to force a polynomial of a certain order to the data then it goes directly here

		print('Forcing polynomial of degree %d' % deg)
		z, SS_res, rank, sing, rc = np.polyfit(x,y, 
				deg, full=True)
		p            = np.poly1d(z)
		sorted_x     = np.sort(x)
		fit_y        = p(sorted_x) #fitting the polynomial to the x coordinates
		dof          = len(x) - len(z)
		numerator    = SS_res
		denominator  = sum(np.array(y)**2)
		fraction     = numerator/denominator
		red_chi_squared  = numerator/dof
		#SS_tot       = np.sum((y - np.mean(y))**2)
		#r_squared    = 1 - ((SS_res / SS_tot) * ((len(y) - 1) 
		#		/ (len(y) - dof - 1)) )

		polydegree   = {'xarray': x,
						    'fit_y': fit_y,
						    'red_chi_squared': red_chi_squared,
						    'degree': deg,
						    'best_fit': z}

		dat.append(polydegree)


#	dat = np.array(dat)
	#to see only r-squared values: dat[:,3]	

	return dat

#x = np.linspace(0,1,100)
#f = 1/4
#y = np.sin(2*np.pi*f*x) + np.random.normal(scale=0.1, size=len(x))
#plt.plot(x,y)

#y_fit = (best_fit(x, y, deg = 8, epsilon = 0.1,force_degree=True)[0]['fit_y'])
#plt.plot(x,y_fit)
#plt.show()


"""
LMC_Winds is a python package written to analyze HST galactic wind data.

Written by TCU Students.
"""

__author__ = 'TCU Students'

from os import path
from sys import version_info

# import warnings
# import matplotlib
# # The native MacOSX backend doesn't work for all:
# with warnings.catch_warnings():
#     warnings.simplefilter("ignore")
#     matplotlib.use('TkAgg')

from .example import *
from .whichLSF import *
from .def_norm_guesses import *
from .def_velocity_conversions import *
from .fitting import *
from .def_read_line_transitions import *
from .def_read_STSCI import *
#from .def_normalize import *
from .Jo_Codes.wakker_read_in import *
#from .April_codes import norm_testing
#from .April_codes.primary_fits_header import *
#from .April_codes.read_STSCI_testing import *
#from .April_codes.simple_normalization import *




#code_dir = path.dirname(path.abspath(__file__))
#with open(path.join(code_dir, 'VERSION')) as version_file:
#   version = version_file.read().strip()
#   if version_info[0] >= 3:
#        v_items = version.split('.')
#        v_items[0] = '3'
#        version = '.'.join(v_items)
#   __version__ = version



import os
from astropy.io import fits
#import def_read_STSCI
#base_directory = os.environ['LMC_WINDS']
#data_directory = os.path.join(base_directory,'Data','HST','dr5','')
#file = ''

def read_STSCI(directory='',file=''):
	"""
    This definition reads in the header information from the STSCI fits files from the ULLYSES program
    Parameters
    ----------
    file :Insert the name of the file from which you want to extract data from
    Returns
    -------
    STSCI_info : dictionary
        This dictionary holds the wavelengths, fluxes, flux errors, target name, ra, dec, instrument, and diffraction grating for the STSCI
        file. You can access a specific value in the dictionary by read_STSCI('hlsp_ullyses_fuse_fuv_bi173_lwrs_dr5_cspec.fits')['RA'] which 
        returns the RA.

    Example on How to Run Code
    --------------------------
    read_STSCI('hlsp_ullyses_fuse_fuv_bi173_lwrs_dr5_cspec.fits') 
    This will return the whole dictionary 

    Written by April Horton 10/17/2022

    """

	STSCI_info = dict()
	spectrum = fits.open(directory+file)
	wl=spectrum[1].data['WAVELENGTH'][0]
	STSCI_info['WAVELENGTH'] = wl

	flux=spectrum[1].data['FLUX'][0]
	STSCI_info['FLUX'] = flux

	flux_err=spectrum[1].data['ERROR'][0]
	STSCI_info['FLUXERR'] = flux_err


	return STSCI_info




#print(read_STSCI('hlsp_ullyses_fuse_fuv_bi173_lwrs_dr5_cspec.fits')['RA'])



#
import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.time import Time

def which_lsf(file_name, rest_wave): 
    """
   This takes the HST spectrum in fits format, reads the header and print the useful information about the resolution, instrument,     and LSF, if applicable, corresponding to the line of interest
    file_name : input fits file    
    rest_wave : rest wavelength in Angstrom corresponding to the lines of interest (e.g., 1302.1685 for O I 1302)
  
    Returns
    -------
    Instrument ID, Filter ID, CENWAVE, Life Position 
    """
     
    #The moves to Lifetime Position 2 (LP2), Lifetime Position 3 (LP3), and Lifetime Position 4 (LP4) 
    #occurred on July 23, 2012, February 9, 2015, and October 2, 2017, respectively
    LP2 = Time('2012-07-23')
    LP3 = Time('2015-02-09')
    LP4 = Time('2017-10-02')
    LP2.format = 'mjd'
    LP3.format = 'mjd'
    LP4.format = 'mjd'
    LP2,LP3,LP4=56131.0,57062.0,58028.0


    hdul=fits.open(file_name)
    field=hdul[2].data
    wave_diff=[]
    for i in field['CENWAVE']:
        j=float(i)-rest_wave
        j=abs(j)
        wave_diff.append(j)  
    min_value=min(wave_diff)
    which=wave_diff.index(min_value)
    
    

    if field['MJD_MID'][which]<LP2:
       life_position = ('LP1')
    if LP2<field['MJD_MID'][which]<LP3:
       life_position = ('LP2')
    if LP3<field['MJD_MID'][which]<LP4:
       life_position = ('LP3')
    if field['MJD_MID'][which]>LP4:
       life_position = ('LP4')

    print ("Instrument, GRATING, CENWAVE, SPECRES, LIFE POSITION")
    print (field['INSTRUMENT'][which], field['DISPERSER'][which], field['CENWAVE'][which],field['SPECRES'][which], life_position)


import os
import numpy as np
#exec(open('def_norm_guesses.py').read())

#from LMC_Winds import def_norm_guesses

# Example:
#     from LMC_Winds import def_norm_guesses
#     import def_norm_guesses
#     target='SK-70D16'
#     def_norm_guesses.norm_guesses(target=target)

#hard coded for directory

# Job to read in the normalization mask zones

def norm_guesses(directory='',target=''):

	direct = os.path.join(directory,target,'') 
	#print(direct)

	norm_dat = 'fit_norm.dat'
	ion_wave, wavemin, wavemax =np.genfromtxt(direct+norm_dat, \
		skip_header=1,unpack=True,dtype="str, float, float")

	norm_mask = []
	for min_wave, max_wave in zip(wavemin,wavemax):
		norm_mask.append([min_wave,max_wave])

	return norm_mask



#print(norm_guesses('SK-70D16')[0])


import matplotlib
from matplotlib import pyplot as plt
import os
import numpy as np
from astropy import constants as const 
from astropy import units as u 
from numpy.polynomial import Polynomial
import def_read_line_transitions
import fitting
import def_normalize
import def_norm_guesses
#exec(open('def_normalize.py').read())

#exec(open('fitting.py').read())
#optional pass for ra and dec just for LMCSR and LSR frames, keep it all helio
#optional vel_span keyword is 1500. 

def normalize(ion_wave_strings,ion_dir, wl, flux, flux_err, velocity_span=1500., plot=True):

	ion_string=ion_wave_strings[0].split('_')[0] 
	num_ions = len(ion_wave_strings)

	norm_info = dict()

	l0 = list(def_read_line_transitions.read_line_transitions(ion_dir,ion_wave_strings)[0].keys())[0]

	for ion_transition in ion_wave_strings:
		#exec('l0 = l0_'+ion_transition) #creating new variable  based on the selected ion_transition
		#exec('norm_mask_ranges = norm_mask_'+ion_transition) #creating a new variable based on the selected ion_transition

		vel_masked = ((wl - l0)/l0) * (const.c.to('km/s')*u.s/u.km)
		vspan_mask = ((vel_masked > -velocity_span) & (vel_masked < velocity_span))

		wl_vel_masked_transition = wl[vspan_mask] #wavelengths inside the corresponding velocity range limits
		norm_info['MASKWL'] = wl_vel_masked_transition

		flux_vel_masked_transition = flux[vspan_mask] #flux values inside the corresponding velocity range limits
		err_vel_masked_transition = flux_err[vspan_mask] #flux err values inside the corresponding velocity range limits

		wl_cont_masked = np.array([]) #holds the wavlengths, fluxes, and errors, that are inbetween the specified velocity span but 
		#have the good continuum regions. essentially masking the continuum where there are no additional absorption features
		flux_cont_masked = np.array([])
		err_cont_masked = np.array([])

		for rang in def_norm_guesses.norm_guesses(directory=directory ,target=target): #may be multiple wavelength ranges to mask out
		#print(rang[0], rang[1])

		#For a list of wavelength ranges, I am selecting the selected range of the continuum to normalize
		
			mask = (wl_vel_masked_transition> rang[0]) & (wl_vel_masked_transition< rang[1])
		
			good_continuum = wl_vel_masked_transition[mask]
			good_flux = flux_vel_masked_transition[mask]
			good_err = err_vel_masked_transition[mask]


			wl_cont_masked = np.append(wl_cont_masked,good_continuum)
			flux_cont_masked = np.append(flux_cont_masked,good_flux) 
			err_cont_masked = np.append(err_cont_masked,good_err) 

		
		auto_order = fitting.best_fit(wl_cont_masked, flux_cont_masked,w=err_cont_masked, deg = 8, epsilon = 0.05,force_degree=False)
		poly = Polynomial.fit(wl_cont_masked,flux_cont_masked,auto_order,w=err_cont_masked)
		
		if plot==True:
			plt.plot(wl_vel_masked_transition,flux_vel_masked_transition, c='k') #whole spectrum
			plt.plot(wl_cont_masked,poly(wl_cont_masked), color='brown')
			plt.plot(wl_cont_masked,flux_cont_masked,c='c') #the continuum

			plt.show()

		norm_flux = flux_vel_masked_transition/poly(wl_vel_masked_transition)
		norm_info['NORMFLUX'] = norm_flux
		norm_flux_error = norm_flux*(np.sqrt((err_vel_masked_transition/flux_vel_masked_transition)**2))
		norm_info['NORMFLUXERR'] = norm_flux_error
		##################################################################
		#here I am going to normalize just the continuum
		norm_continuum = flux_cont_masked/poly(wl_cont_masked)
		#norm_continuum_error = err_cont_masked/poly(wl_cont_masked)
		if plot == True:
			plt.plot(wl_vel_masked_transition, norm_flux) # these could be saved in a table as they are the x and y values of the normalized spectra
			plt.plot(wl_vel_masked_transition,norm_flux*0+1, c='k')
			plt.show()

	
	return norm_info

		

#print(norm_dictionary['NORMFLUX'])







from setuptools import setup 
#from Cython.Build import cythonize

setup(
    name="LMC_Winds",
    version="0.0.2",
    description="Python package for analyzing HST data related to LMC",
    author="TCU Students",
    license="None",
    author_email="suraj.poudel@tcu.edu",
    #url="https:",
    packages=['LMC_Winds'])


#exec(open('def_velocity_conversions.py').read())
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.coordinates import ICRS, Galactic, FK4, FK5 
from astropy.coordinates.representation import UnitSphericalRepresentation
def vhelio_to_vlsr(coord, vhelio=0*u.km/u.s, lsr_definition='kinematic'): #coord is the coordinates of the background flashlight (the RA and DEC in deg)
    """
    this definition is from the work of erikutils. https://github.com/eteq/erikutils/blob/master/erikutils/velocities.py
    Convert heliocentric radial velocity to Local Standard of Rest radial velocity
    Parameters
    ----------
    coord : SkyCoord
        The direction at which to compute the conversion factor.
    vhelio : Quantity with velocity units
        The heliocentric radial velocity.  Should be a scalar quantity or match the shape of `coord`.
    lsr_definition : str or Quantity with velocity units
        The definition of LSR to assume.  This can be one of three options:
        * 'kinematic' : 20 km/s towards 18h,+30 deg (at 1900)
        * 'dynamical' : IAU definition of (9, 12, 7) km/sec in Galactic cartesian coordinates
        * A length-3 vector with the U, V, and W components (velocity along the galactic x, y, and z axes)
          to assume for the LSR.
    Returns
    -------
    vlsr : Quantity
        The velocity in the Local Standard of Rest frame.
    """
    if lsr_definition == 'kinematic':
        direction = SkyCoord('18h', '30d', frame='fk5', equinox='J1900') #definition of kinematic reference frame
        velocity = 20*u.km/u.s #definition of kinematic reference frame
        uvw_lsr = direction.galactic.cartesian.xyz * velocity #puts the direction in galactic frame and then in the cartesian
        #The xyz says remember all three coordinates, the x,y and z in the cartesian frame
    elif lsr_definition == 'dynamical':
        uvw_lsr = (9, 12, 7)*u.km/u.s
    else:
        uvw_lsr = lsr_definition

    # the unitspherical conversion ensures that the resulting cartesian is a *unit* vector
    usr = coord.galactic.represent_as(UnitSphericalRepresentation) #UnitSphericalRepresentation represents the points on a unit sphere
    cart = usr.to_cartesian()

    vlsr_vector = (cart.xyz.T*uvw_lsr).T #anytime you see the .xyz it is just return the x,y, and z components of the vector
    #the .T is the transpose of the numpy array. Apparently, the .T is np.transpose()
    return vhelio + np.sum(vlsr_vector)

def vlsr_to_lmcsr(coord, VLSR):
    """
    Converts from the Local Standard of Reset to LMC Standard of Rest reference frame
    Parameters
    ----------
    coord : SkyCoord
        The the RA and DEC of the sightline in degrees
    VLSR : Quantity with units of km/s

    Returns
    -------
    vlmcsr : the velocity in the LMCSR reference frame

    Notes
    ------
    This conversion to LMCSR is using a modified version of equation 3 in Ciampa et al. 2021. In Ciampa et al. 2021, the central velocity offset
    is found from the following equation: delta_vlmcsr = 262.55-3.25(glon-280)+3.66(glat-33). Then the central velocity offset was added to the local standard
    of rest velocity. The corrections for this code include the following new equation: (262.55-3.25*(glon-280.)+3.66*(glat-(-33.))) and this offset is 
    subtracted from the local standard of rest velocity.

    Written by April Horton 05/17/2022
    """
    c = coord #represents an ICRS (RA and DEC sky position), coord should be a SkyCoord object SkyCoord(ra=10.625*u.degree, dec=41.2*u.degree, frame='icrs')
    #To get the coordinate in Galactic Frame, use the .galactic attribute
    galactic_c = c.galactic 
    galactic_l = galactic_c.l.value
    galactic_b = galactic_c.b.value
    # Talking point in my paper
    central_v_offset = 262.55-3.25*(galactic_l-280)+3.66*(galactic_b-(-33)) # Modified equation 3 from Ciampa et al. 2021, The modifications include subtracting (-33) from the galactic longitude 
    #central_v_offset = 1293.22 - 3.25*(galactic_l)+3.66*(galactic_b) #this equation is from the screenshot (Drew_original_equation.png) in my journal articles directory. It has what originally Drew had in his paper before the referee comments
    
    vlmcsr = VLSR - central_v_offset #this is also modified from Ciampa et al. 2021. We subtract the central velocity offset from the LSR velocity, still subtract with the equation from the screenshot
    #print(central_v_offset)
    
    return (vlmcsr)
